import java.util.Random;

public class Test {

    private static Random generator = new Random();

    private static int nextInt() {
        return generator.nextInt(100);
    }

    private static boolean nextBool() {
        return generator.nextBoolean();
    }

    private static double nextDouble() {
        return generator.nextDouble() * 100.0;
    }

    public static void main(String[] argv) {

        Foo[] foos = new Foo[10];

        for (int i=0; i < foos.length; i++) {
            if (i % 2 == 0) {
                foos[i] = new Bar(new BarSpec(nextInt(), nextBool()));
            } else {
                foos[i] = new Baz(new BazSpec(nextInt(), nextDouble()));
            }
        }

        System.out.println("============================================");
        System.out.println("doSomething() in Foo/Bar/Baz is polymorphic.");
        System.out.println("toString() in FooSpec/BarSpec/BazSpec is too.");
        System.out.println();
        System.out.println("That's why invoking doSomething() on a Foo object");
        System.out.println("does the right thing. See below:");
        System.out.println();
        for (Foo foo : foos) {
            foo.doSomething();
        }
        System.out.println();

        BarSpec bar_spec = new BarSpec(nextInt(), nextBool());
        BazSpec baz_spec = new BazSpec(nextInt(), nextDouble());

        // The next bit of code shows that our matches methods in
        // BarSpec and BazSpec do not override the matches method in
        // FooSpec. They overload it and so we don't get polymorphic
        // behavior.
        System.out.println("========================================================");
        System.out.println("matches() in FooSpec/BarSpec/BazSpec is NOT polymorphic.");
        System.out.println();
        System.out.println("That's why invoking matches() on a FooSpec object");
        System.out.println("does not invoke matches in the subclass.");
        System.out.println();
        for (Foo foo : foos) {
            FooSpec s = foo.getSpec();
            System.out.println("s matches: " + s.matches(bar_spec));
        }
        System.out.println();

        // Because they do not override matches, in order to get what we
        // want, we have to do a bit more work

        System.out.println("==============================================");
        System.out.println("In order to invoke matches in the subclass, we");
        System.out.println("have to cast our variable to the appropriate");
        System.out.println("type.");
        System.out.println();
        for (Foo foo : foos) {
            FooSpec s = foo.getSpec();
            if (s instanceof BarSpec) {
              BarSpec b_spec = (BarSpec)s;
              System.out.println("b_spec matches: " + b_spec.matches(bar_spec));
            } else {
              BazSpec b_spec = (BazSpec)s;
              System.out.println("b_spec matches: " + b_spec.matches(baz_spec));
            }
        }
        System.out.println();
    }
}
